@@ -45,7 +45,8 @@ timeouts = {'short': 100, 'copy': 300, 'install': 3000, 'test': 10000,
 
 class Testbed:
     def __init__(self, vserver_argv, output_dir, user,
-                 setup_commands=[], add_apt_pockets=[], copy_files=[]):
+                 setup_commands=[], setup_commands_boot=[], add_apt_pockets=[],
+                 copy_files=[]):
         self.sp = None
         self.lastsend = None
         self.scratch = None
@@ -60,6 +61,7 @@ class Testbed:
         self.install_tmp_env = []
         self.user = user
         self.setup_commands = setup_commands
+        self.setup_commands_boot = setup_commands_boot
         self.add_apt_pockets = add_apt_pockets
         self.copy_files = copy_files
         self.initial_kernel_version = None
@@ -186,6 +188,21 @@ class Testbed:
             if m:
                 self.cpu_flags = m.group(2)
 
+        xenv = ['AUTOPKGTEST_IS_SETUP_BOOT_COMMAND=1']
+        if self.user:
+            xenv.append('AUTOPKGTEST_NORMAL_USER=' + self.user)
+            xenv.append('ADT_NORMAL_USER=' + self.user)
+
+        for c in self.setup_commands_boot:
+            rc = self.execute(['sh', '-ec', c], xenv=xenv, kind='install')[0]
+            if rc:
+                # setup scripts should exit with 100 if it's the package's
+                # fault, otherwise it's considered a transient testbed failure
+                if rc == 100:
+                    self.badpkg('testbed boot setup commands failed with status 100')
+                else:
+                    self.bomb('testbed boot setup commands failed with status %i' % rc)
+
     def _opened(self, pl):
         self.scratch = pl[0]
         self.deps_installed = []
@@ -255,7 +272,7 @@ class Testbed:
         # create apt sources for --apt-pocket
         for pocket in self.add_apt_pockets:
             pocket = pocket.split('=', 1)[0]  # strip off package list
-            script = '''sed -rn 's/^(deb|deb-src) +(\[.*\] *)?([^ ]*(ubuntu.com|debian.org|ftpmaster|file:\/\/\/tmp\/testarchive)[^ ]*) +([^ -]+) +(.*)$/\\1 \\2\\3 \\5-%s \\6/p' /etc/apt/sources.list `ls /etc/apt/sources.list.d/*.list 2>/dev/null|| true)` > /etc/apt/sources.list.d/%s.list; for retry in 1 2 3; do apt-get --no-list-cleanup -o Dir::Etc::sourcelist=/etc/apt/sources.list.d/%s.list -o Dir::Etc::sourceparts=/dev/null update 2>&1 && break || sleep 15; done''' % (pocket, pocket, pocket)
+            script = '''sed -rn 's/^(deb|deb-src) +(\[.*\] *)?([^ ]*(ubuntu.com|debian.org|ftpmaster|file:\/\/\/tmp\/testarchive)[^ ]*) +([^ -]+) +(.*)$/\\1 \\2\\3 \\5-%s \\6/p' /etc/apt/sources.list `ls /etc/apt/sources.list.d/*.list 2>/dev/null|| true` > /etc/apt/sources.list.d/%s.list; for retry in 1 2 3; do apt-get --no-list-cleanup -o Dir::Etc::sourcelist=/etc/apt/sources.list.d/%s.list -o Dir::Etc::sourceparts=/dev/null update 2>&1 && break || sleep 15; done''' % (pocket, pocket, pocket)
             self.check_exec(['sh', '-ec', script])
 
         # create apt pinning for --apt-pocket with package list
@@ -313,7 +330,7 @@ class Testbed:
             self._opened(pl)
         self.modified = False
 
-    def install_deps(self, deps_new, recommends):
+    def install_deps(self, deps_new, recommends, shell_on_failure=False):
         '''Install dependencies into testbed'''
         adtlog.debug('install_deps: deps_new=%s, recommends=%s' % (deps_new, recommends))
 
@@ -321,7 +338,7 @@ class Testbed:
         self.recommends_installed = recommends
         if not deps_new:
             return
-        self.satisfy_dependencies_string(', '.join(deps_new), 'install-deps', recommends)
+        self.satisfy_dependencies_string(', '.join(deps_new), 'install-deps', recommends, shell_on_failure=shell_on_failure)
 
     def needs_reset(self):
         # show what caused a reset
@@ -351,28 +368,28 @@ class Testbed:
                       format_exception_only(type, value))
 
     def expect(self, keyword, nresults):
-        l = self.sp.stdout.readline()
-        if not l:
+        line = self.sp.stdout.readline()
+        if not line:
             self.bomb('unexpected eof from the testbed')
-        if not l.endswith('\n'):
+        if not line.endswith('\n'):
             self.bomb('unterminated line from the testbed')
-        l = l.rstrip('\n')
-        adtlog.debug('got reply from testbed: ' + l)
-        ll = l.split()
+        line = line.rstrip('\n')
+        adtlog.debug('got reply from testbed: ' + line)
+        ll = line.split()
         if not ll:
             self.bomb('unexpected whitespace-only line from the testbed')
         if ll[0] != keyword:
             if self.lastsend is None:
                 self.bomb("got banner `%s', expected `%s...'" %
-                          (l, keyword))
+                          (line, keyword))
             else:
                 self.bomb("sent `%s', got `%s', expected `%s...'" %
-                          (self.lastsend, l, keyword))
+                          (self.lastsend, line, keyword))
         ll = ll[1:]
         if nresults is not None and len(ll) != nresults:
             self.bomb("sent `%s', got `%s' (%d result parameters),"
                       " expected %d result parameters" %
-                      (self.lastsend, l, len(ll), nresults))
+                      (self.lastsend, line, len(ll), nresults))
         return ll
 
     def command(self, cmd, args=(), nresults=0, unquote=True):
@@ -443,6 +460,7 @@ class Testbed:
         adtlog.debug('testbed command exited with code %i' % proc.returncode)
 
         if proc.returncode in (254, 255):
+            self.command('auxverb_debug_fail')
             self.bomb('testbed auxverb failed with exit code %i' % proc.returncode)
 
         return (proc.returncode, out, err)
@@ -505,6 +523,7 @@ Description: satisfy autopkgtest test dependencies
                                           '--assume-yes --fix-broken '
                                           '-o APT::Status-Fd=3 '
                                           '-o APT::Install-Recommends=%s '
+                                          '-o Dpkg::Options::=--force-confnew '
                                           '-o Debug::pkgProblemResolver=true 3>&2 2>&1' %
                                           (' '.join(self.eatmydata_prefix), recommends)],
                                          kind='install', stderr=subprocess.PIPE)
@@ -538,6 +557,8 @@ Description: satisfy autopkgtest test dependencies
                     self.check_exec(['/bin/sh', '-ec', 'rm /etc/apt/preferences.d/autopkgtest-*-' + pocket])
                     continue
 
+                if shell_on_failure:
+                    self.run_shell()
                 self.badpkg('Test dependencies are unsatisfiable. A common reason is '
                             'that your testbed is out of date with respect to the '
                             'archive, and you need to use a current testbed or run '
@@ -979,7 +1000,7 @@ fi
             test_cmd = "bash -ec '%s'" % test.command
 
         script += 'touch %(o)s %(e)s; ' \
-                  '%(t)s 2> >(tee -a %(e)s >&2) > >(tee -a %(o)s); ' \
+                  '%(t)s 2> >(tee -a %(e)s >&2) > >(tee -a %(o)s);' \
                   % {'t': test_cmd, 'o': so.tb, 'e': se.tb}
 
         if 'needs-root' not in test.restrictions and self.user is not None:
@@ -1055,9 +1076,21 @@ fi
         adtlog.debug('testbed executing test finished with exit status %i' % rc)
 
         # copy stdout/err files to host
-        so.copyup()
-        se.copyup()
-        se_size = os.path.getsize(se.host)
+        try:
+            so.copyup()
+            se.copyup()
+            se_size = os.path.getsize(se.host)
+        except adtlog.TestbedFailure:
+            if timeout:
+                # if the test timed out, it's likely that the test destroyed
+                # the testbed, so ignore this and call it a failure
+                adtlog.warning('Copying up test output timed out, ignoring')
+                se_size = 0
+                so.host = None
+                se.host = None
+            else:
+                # smells like a tmpfail
+                raise
 
         # avoid mixing up stdout (from report) and stderr (from logging) in output
         sys.stdout.flush()
@@ -1080,6 +1113,11 @@ fi
         sys.stdout.flush()
         sys.stderr.flush()
 
+        # skip the remaining processing if the testbed got broken
+        if se.host is None:
+            adtlog.debug('Skipping remaining log processing and testbed restore after timeout')
+            return
+
         if os.path.getsize(so.host) == 0:
             # don't produce empty -stdout files in --output-dir
             so.autoclean = True
