@@ -42,7 +42,7 @@ import adtlog
 
 known_restrictions = ['rw-build-tree', 'breaks-testbed', 'needs-root',
                       'build-needed', 'allow-stderr', 'isolation-container',
-                      'isolation-machine', 'needs-recommends']
+                      'isolation-machine', 'needs-recommends', 'needs-reboot']
 
 
 class Unsupported(Exception):
@@ -161,6 +161,12 @@ class Test:
                               'Test needs root on testbed which is not '
                               'available')
 
+        if 'needs-reboot' in self.restrictions and \
+           'reboot' not in caps:
+            raise Unsupported(self.name,
+                              'Test needs to reboot testbed but testbed does '
+                              'not provide reboot capability')
+
 #
 # Parsing for Debian source packages
 #
@@ -181,18 +187,18 @@ def parse_rfc822(path):
     # filter out comments, python-debian doesn't do that
     # (http://bugs.debian.org/743174)
     lines = []
-    for l in f:
+    for line in f:
         # completely ignore ^# as that breaks continuation lines
-        if l.startswith('#'):
+        if line.startswith('#'):
             continue
         # filter out comments which don't start on first column (Debian
         # #743174); entirely remove line if all that's left is whitespace, as
         # that again breaks continuation lines
-        if '#' in l:
-            l = l.split('#', 1)[0]
-            if not l.strip():
+        if '#' in line:
+            line = line.split('#', 1)[0]
+            if not line.strip():
                 continue
-        lines.append(l)
+        lines.append(line)
     f.close()
 
     for p in debian.deb822.Deb822.iter_paragraphs(lines):
@@ -383,7 +389,25 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
         try:
             restrictions = record.get('Restrictions', '').replace(
                 ',', ' ').split()
-            features = record.get('Features', '').replace(',', ' ').split()
+
+            feature_test_name = None
+            features = []
+            record_features = record.get('Features', '').replace(
+                ',', ' ').split()
+            for feature in record_features:
+                details = feature.split('=', 1)
+                if details[0] != 'test-name':
+                    features.append(feature)
+                    continue
+                if len(details) != 2:
+                    # No value, i.e. a bare 'test-name'
+                    raise InvalidControl(
+                        '*', 'test-name feature with no argument')
+                if feature_test_name is not None:
+                    raise InvalidControl(
+                        '*', 'only one test-name feature allowed')
+                feature_test_name = details[1]
+                features.append(feature)
 
             if 'Tests' in record:
                 test_names = record['Tests'].replace(',', ' ').split()
@@ -394,6 +418,9 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
                 if 'Test-command' in record:
                     raise InvalidControl('*', 'Only one of "Tests" or '
                                          '"Test-Command" may be given')
+                if feature_test_name is not None:
+                    raise InvalidControl(
+                        '*', 'test-name feature incompatible with Tests')
                 test_dir = record.get('Tests-directory', 'debian/tests')
 
                 _debian_check_unknown_fields(test_names[0], record)
@@ -408,8 +435,11 @@ def parse_debian_source(srcdir, testbed_caps, testbed_arch, control_path=None,
                                                 record.get('Depends', '@'),
                                                 srcdir,
                                                 testbed_arch)
-                command_counter += 1
-                name = 'command%i' % command_counter
+                if feature_test_name is None:
+                    command_counter += 1
+                    name = 'command%i' % command_counter
+                else:
+                    name = feature_test_name
                 _debian_check_unknown_fields(name, record)
                 test = Test(name, None, command, restrictions, features,
                             depends, [], [])
